from __future__ import annotations

import heapq
import random
import uuid
from fractions import Fraction
from functools import reduce
from typing import Any, Mapping, Optional, Sequence, Union

import aiger
import aiger_bv as BV
import aiger_discrete
import attr
import funcy as fn
from aiger_discrete import FiniteFunc


Prob = Union[float, Fraction]
Distribution = Mapping[Any, Prob]


def coin_gadget(name: str,
                dist: Distribution,
                func: FiniteFunc,
                tree_encoding: bool = True) -> PCirc:
    """Return a Probabilistic Circuit representing a distribution.

    - chain encoding. O(|support|)
    - tree encoding. O(log|support|)
    """
    # Transform support of dist to UnsignedBVExpr using encoding.
    size = func.circ.imap[name].size
    encode = func.input_encodings \
                 .get(name, aiger_discrete.Encoding()) \
                 .encode
    dist = ((p, BV.uatom(size, encode(v))) for v, p in dist.items())

    # Create priority queue.
    cost = len if tree_encoding else (lambda x: -len(x))
    queue = [(cost(elem), elem) for elem in dist]

    coins = []
    while len(queue) > 1:
        cost1, (weight1, expr1) = heapq.heappop(queue)
        cost2, (weight2, expr2) = heapq.heappop(queue)

        # Create new coin input.
        coin = BV.uatom(1, None)
        bias = weight1 / (weight1 + weight2)
        coins.append((fn.first(coin.inputs), bias))

        # Merge and add back to queue.
        expr12 = BV.ite(coin, expr1, expr2)
        cost12 = cost1 + cost2  # (-x + -y) == -(x + y).
        weight12 = weight1 + weight2
        heapq.heappush(queue, (cost12, (weight12, expr12)))
    assert len(queue) == 1
    _, (weight, expr) = queue[0]

    expr = expr.bundle_inputs(order=[c for c, _ in coins]) \
               .with_output(name)
    assert len(expr.inputs) == 1
    coins_id = fn.first(expr.inputs)
    biases = tuple(bias for _, bias in coins)
    return PCirc(expr.aigbv, coins_id=coins_id, coin_biases=biases)


def coin_gadgets(dist_map: Mapping[str, Distribution],
                 func: FiniteFunc,
                 tree_encoding: bool = True) -> PCirc:
    """Return a Probabilistic Circuit representing a product distribution."""
    gadgets = (
        coin_gadget(k, v, func, tree_encoding) for k, v in dist_map.items()
    )
    return reduce(PCirc.__or__, gadgets)


def to_finite_func(circ) -> FiniteFunc:
    if isinstance(circ, FiniteFunc):
        return circ
    return aiger_discrete.from_aigbv(circ.aigbv)


def merge_pcirc_coins(circ, left: PCirc, right: PCirc, coins_id: str):
    if not left.has_coins and not right.has_coins:
        biases = ()
    elif left.has_coins and not right.has_coins:
        circ <<= BV.uatom(left.num_coins, coins_id) \
                   .with_output(left.coins_id).aigbv
        biases = left.coin_biases
    elif right.has_coins and not left.has_coins:
        circ <<= BV.uatom(right.num_coins, coins_id) \
                   .with_output(right.coins_id).aigbv
        biases = right.coin_biases
    else:
        coins = BV.uatom(left.num_coins + right.num_coins, coins_id)

        relabeler = coins[:left.num_coins].with_output(left.coins_id).aigbv \
                  | coins[left.num_coins:].with_output(right.coins_id).aigbv  # noqa: E127, E501
        circ <<= relabeler
        biases = tuple(left.coin_biases) + tuple(right.coin_biases)
    return circ, biases


def sample_coins(coin_biases: Sequence[float]) -> int:
    """Return integer where bits bias towards 1 as described in coin_biases."""
    result = 0
    for i, bias in enumerate(coin_biases):
        result |= int(random.random() < bias) << i
    return result


@attr.s(frozen=True, auto_attribs=True)
class PCirc:
    """Wrapper around AIG representing a function with some random inputs."""
    circ: FiniteFunc = attr.ib(converter=to_finite_func)
    coin_biases: Sequence[Prob]           # Bias of each coin flips.
    coins_id: str = "##coins"             # Input reservered for coin flips.

    def __attrs_post_init__(self):
        if not self.has_coins:
            return
        circ: BV.AIGBV = self.circ.circ
        if self.coins_id not in circ.imap:
            raise ValueError("Underlying circuit doesn't have coins input.")
        if circ.imap[self.coins_id].size != len(self.coin_biases):
            raise ValueError("Underlying circuit doesn't have enough coins.")

    @property
    def has_coins(self): return len(self.coin_biases) > 0

    @property
    def inputs(self): return self.circ.inputs - {self.coins_id}

    @property
    def outputs(self): return self.circ.outputs

    @property
    def latches(self): return self.circ.latches

    @property
    def latch2init(self): return self.circ.latch2init

    @property
    def aig(self): return self.circ.aig

    @property
    def aigbv(self): return self.circ.aigbv

    @property
    def num_coins(self): return len(self.coin_biases)

    @property
    def imap(self):
        return self.circ.imap.omit([self.coins_id])

    @property
    def omap(self):
        return self.circ.omap.omit([self.coins_id])

    def assume(self, aigbv_like) -> PCirc:
        """Return Probabilistic Circuit with new assumption over the inputs."""
        return attr.evolve(self, circ=self.circ.assume(aigbv_like))

    def with_coins_id(self, name=None):
        if name is None:
            name = str(uuid.uuid1())
        circ = self.circ
        if self.has_coins:
            circ = self.circ['i', {self.coins_id: name}]
        return attr.evolve(self, circ=circ, coins_id=name)

    def __rshift__(self, other) -> PCirc:
        other = canon(other)
        circ = self.circ >> other.circ
        circ, biases = merge_pcirc_coins(circ, self, other, self.coins_id)
        return PCirc(circ, coins_id=self.coins_id, coin_biases=biases)

    def __lshift__(self, other) -> PCirc:
        return canon(other) >> self

    def __or__(self, other) -> PCirc:
        other = canon(other)
        circ = self.circ | other.circ
        circ, biases = merge_pcirc_coins(circ, self, other, self.coins_id)
        return PCirc(circ, coins_id=self.coins_id, coin_biases=biases)

    def __call__(self, inputs, latches=None):
        inputs = dict(inputs)
        if self.has_coins:
            inputs.setdefault(self.coins_id, sample_coins(self.coin_biases))
        return self.circ(inputs=inputs, latches=latches)

    def __getitem__(self, others) -> PCirc:
        circ = self.circ[others]
        kind, relabels = others
        if (kind == 'i') and (self.coins_id in relabels):
            raise ValueError("Use with_coins_id to relabel coins.")
        return attr.evolve(self, circ=circ)

    def loopback(self, *wirings) -> PCirc:
        inputs = self.inputs
        assert all(w['input'] in inputs for w in wirings)
        circ = self.circ.loopback(*wirings)
        return attr.evolve(self, circ=circ)

    def unroll(self,
               horizon, *,
               init=True,
               omit_latches=True,
               only_last_outputs=False) -> PCirc:
        # Unroll underlying circuit.
        circ = self.circ.unroll(
            horizon=horizon, init=init, omit_latches=omit_latches,
            only_last_outputs=only_last_outputs,
        )

        if not self.has_coins:
            return attr.evolve(self, circ=circ)

        # Merge timed coin sequences into a single input.
        coins = BV.uatom(self.num_coins * horizon, self.coins_id)
        relabeler = BV.aig2aigbv(aiger.empty())
        for time in range(horizon):
            name = f"{self.coins_id}##time_{time}"
            assert name in circ.inputs

            start = time * self.num_coins
            end = start + self.num_coins
            relabeler |= coins[start:end].with_output(name).aigbv
        circ <<= relabeler
        biases = self.coin_biases * horizon

        return PCirc(circ, coins_id=self.coins_id, coin_biases=biases)

    def randomize(self, dist_map: Mapping[str, Distribution]) -> PCirc:
        """Apply distributions in dist_map to corresponding inputs."""
        circ = BV.aig2aigbv(aiger.empty())
        for name in dist_map.keys():
            size = self.circ.circ.imap[name].size  # TODO: propogate imap.
            circ |= BV.identity_gate(size, name)

        func = aiger_discrete.from_aigbv(
            circ=circ,
            input_encodings=self.circ.input_encodings,
            output_encodings=self.circ.output_encodings,
        )
        return pcirc(func, dist_map) >> self

    simulator = aiger_discrete.FiniteFunc.simulator
    simulate = aiger_discrete.FiniteFunc.simulate


def pcirc(func,
          dist_map: Optional[Mapping[str, Distribution]] = None,
          tree_encoding: bool = False) -> PCirc:
    """Lift Discrete Function to a probilistic circuit."""
    func = to_finite_func(func)
    if dist_map is None:
        return PCirc(circ=func, coin_biases=())

    gadgets = coin_gadgets(dist_map, func, tree_encoding)
    return gadgets >> func


def canon(circ) -> PCirc:
    if not isinstance(circ, PCirc):
        circ = PCirc(circ, coin_biases=())
    return circ.with_coins_id()


__all__ = ['PCirc', 'pcirc']



from functools import reduce

import aiger_bv as BV
import aiger_discrete as D
import funcy as fn
import networkx as nx
import numpy as np

import aiger_coins as C


FALSE = 0b01
TRUE = 0b10


def onehot_gadget(circ):
    output = fn.first(circ.outputs)
    sat = BV.uatom(1, output)
    false, true = BV.uatom(2, FALSE), BV.uatom(2, TRUE)
    return BV.ite(sat, true, false).aigbv


def prob(circ, *, log=False):
    """Return probability that the output of circ is True given a valid run."""
    # Note: circ represents Pr(query & valid).
    # Want: Pr(query | valid) = Pr(query & valid) / Pr(valid)
    #
    # Can answer Pr(valid) by creating another circuit such that:
    #
    #      query' = valid, and valid' = True.
    #
    valid_test = circ.circ.aigbv.cone(circ.circ.valid_id)
    valid_test = C.PCirc(
        circ=D.from_aigbv(valid_test),
        coin_biases=circ.coin_biases,
        coins_id=circ.coins_id,
    )
    result = _prob(circ) - _prob(valid_test)
    return result if log else np.exp(result)


def _prob(circ):
    from aiger_discrete.mdd import to_mdd
    from mdd.nx import to_nx

    if circ.inputs:
        raise ValueError("All inputs must be randomized.")

    if len(circ.outputs) != 1:
        raise ValueError("Only support querying probability of single output.")

    # Make coins different variables for MDD.
    coins = (BV.uatom(1, f"coin_{i}") for i in range(circ.num_coins))
    coin_blaster = reduce(lambda x, y: x.concat(y), coins)
    coin_blaster = coin_blaster.with_output(circ.coins_id) \
                               .aigbv

    #            Seperate coins          MDD expects 1-hot output
    query = (circ.circ << coin_blaster) >> onehot_gadget(circ)

    # View graph of MDD as circuit over LogSumExp
    graph = to_nx(to_mdd(query))
    biases = {f"coin_{i}": bias for i, bias in enumerate(circ.coin_biases)}

    lprobs = {}
    for node in nx.topological_sort(graph.reverse()):
        val = graph.nodes[node]['label']
        if val == TRUE:
            lprobs[node] = 0
        elif val == FALSE:
            lprobs[node] = -float('inf')
        elif graph.out_degree(node) == 1:
            child, *_ = graph.neighbors(node)
            lprobs[node] = lprobs[child]
        else:
            left, right = graph.neighbors(node)
            # Swap if polarity switched.
            if graph.edges[node, right]['label']({val: True})[0]:
                right, left = left, right

            bias = float(biases[val])
            log_biases = np.log(np.array([bias, 1 - bias]))
            kid_lprobs = np.array([lprobs[left], lprobs[right]])

            # Compute average likelihood in log scale.
            lprobs[node] = np.logaddexp(*(log_biases + kid_lprobs))

    return lprobs[node]


def coins_preimage(pcirc, *,
                   inputs=None,
                   outputs=None,
                   latchins=None,
                   latchouts=None) -> BV.UnsignedBVExpr:
    if not pcirc.has_coins:
        return BV.uatom(1, 1)

    # Default IO.
    if inputs is None:
        inputs = {}
    if outputs is None:
        outputs = {}
    if latchouts is None:
        latchouts = {}
    if latchins is None:
        latchins = pcirc.latch2init

    # Check interface.
    assert set(inputs) == pcirc.inputs
    assert set(outputs) == pcirc.outputs
    assert set(latchouts) == pcirc.latches
    assert set(latchins) == pcirc.latches

    # Encode IO.
    mappings = [
        (inputs, pcirc.circ.input_encodings),
        (outputs, pcirc.circ.output_encodings),
    ]
    for val_map, enc_map in mappings:
        val_map.update({k: e.encode(val_map[k]) for k, e in enc_map.items()})

    # Cutlatches and add to i/o.
    circ, relabels = pcirc.aigbv.cutlatches()
    relabels = {k: v for k, (v, _) in relabels.items()}  # Drop init info.
    inputs.update({relabels[k]: v for k, v in latchins.items()})
    outputs.update({relabels[k]: v for k, v in latchouts.items()})

    # Fix inputs.
    srcs = (
        BV.uatom(circ.imap[k].size, v).with_output(k).aigbv
        for k, v in inputs.items()
    )
    circ <<= reduce(lambda x, y: x | y, srcs)
    assert len(circ.inputs) == int(pcirc.has_coins)

    # Test for outputs.
    omap = circ.omap
    tests = (BV.uatom(omap[k].size, k) == v for k, v in outputs.items())
    circ >>= reduce(lambda x, y: x & y, tests).aigbv
    assert len(circ.outputs - {pcirc.circ.valid_id}) == bool(circ.outputs)

    # Test for valid.
    if len(circ.outputs) == 2:  # Force valid.
        left, right = (BV.uatom(1, o) for o in circ.outputs)
        circ >>= (left & right).aigbv

    assert len(circ.outputs) == 1
    return BV.UnsignedBVExpr(circ)


def find_coins(pcirc, *,
               inputs=None,
               outputs=None,
               latchins=None,
               latchouts=None):
    from aiger_sat.sat_bv import solve
    expr = coins_preimage(
        pcirc, inputs=inputs, outputs=outputs,
        latchins=latchins, latchouts=latchouts
    )
    model = solve(expr)
    return {k: BV.decode_int(v, signed=False) for k, v in model.items()}


__all__ = ['prob', 'coins_preimage', 'find_coins']



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiger_coins']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=20.0.0,<21.0.0',
 'funcy>=1.13,<2.0',
 'py-aiger-bv>=4.4.0,<5.0.0',
 'py-aiger-discrete>=0.1.2,<0.2.0',
 'py-aiger>=6.1.1,<7.0.0']

extras_require = \
{'bdd': ['py-aiger-bdd>=3.0.0,<4.0.0',
         'mdd>=0.3.4,<0.4.0',
         'numpy>=1.19.4,<2.0.0'],
 'sat': ['py-aiger-sat>=3.0.4,<4.0.0']}

setup_kwargs = {
    'name': 'py-aiger-coins',
    'version': '3.3.2',
    'description': 'Library for creating circuits that encode discrete distributions.',
    'long_description': "# py-aiger-coins\n\n**warning** 3.0.0 and greater are a **major** rewrite of this code\nbase. I am trying to port most of the useful features.\n\n\n[![Build Status](https://cloud.drone.io/api/badges/mvcisback/py-aiger-coins/status.svg)](https://cloud.drone.io/mvcisback/py-aiger-coins)\n[![codecov](https://codecov.io/gh/mvcisback/py-aiger-coins/branch/master/graph/badge.svg)](https://codecov.io/gh/mvcisback/py-aiger-coins)\n[![Updates](https://pyup.io/repos/github/mvcisback/py-aiger-coins/shield.svg)](https://pyup.io/repos/github/mvcisback/py-aiger-coins/)\n[![PyPI version](https://badge.fury.io/py/py-aiger-coins.svg)](https://badge.fury.io/py/py-aiger-coins)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n\n\nLibrary for creating circuits that encode discrete distributions and\nMarkov Decision Processes. The name comes from the random bit model of\ndrawing from discrete distributions using coin flips.\n\n<!-- markdown-toc start - Don't edit this section. Run M-x markdown-toc-refresh-toc -->\n**Table of Contents**\n\n- [py-aiger-coins](#py-aiger-coins)\n- [Install](#install)\n- [Usage](#usage)\n\n<!-- markdown-toc end -->\n\n\n# Install\n\nTo install this library run:\n\n`$ pip install py-aiger-coins`\n\nNote that to actually compute probabilities, one needs to install with the bdd option.\n\n`$ pip install py-aiger-coins[bdd]`\n\nFor developers, note that this project uses the\n[poetry](https://poetry.eustace.io/) python package/dependency\nmanagement tool. Please familarize yourself with it and then run:\n\n`$ poetry install`\n\n# Usage\n\n`py-aiger-coins` extends the standard `py-aiger-bv` and\n`py-aiger-discrete` abstractions to allow for certain bits to be set\nvia **biased** coins.\n\nThe library centers around the `PCirc` object. The easiest way to use\n`py-aiger-coins` is to throught the `pcirc` function.\n\n\n```python\nimport aiger_bv as BV\nimport aiger_coins as C\n\n\nexpr1 = BV.uatom(2, 'x') + BV.uatom(2, 'y')\n\n# Create distribution over bits.\nmy_pcirc = C.pcirc(expr1) \\\n            .randomize({'x': {0: 1/6, 1: 2/6, 2: 3/6}})\n\nassert my_pcirc.outputs == {expr1.output}\n\n# This probablistic circuit uses 2 biased coins to represent this input.\nassert my_pcirc.num_coins == 2\nassert my_pcirc.coin_biases == (1/3, 1/2)\n\n# 'x' input is replaced with coinflips.\nassert my_pcirc.inputs == {'y'}\n\n# Underlying circuit now has a single input representing coin inputs.\nunderlying_circ = my_pcirc.circ\n\nassert underlying_circ.inputs == {'y', my_pcirc.coins_id}\n```\n\nNote that `aiger_coins.PCirc` implements the same API as `aiger_bv`\nand `aiger_coins`.\n\n## Sequential Circuit API\n\nFor example, sequential and parallel composition allow combining\nprobablistic circuits.\n\n```python\nincr = (BV.uatom(2, 'z') + 1)\nadder = (BV.uatom(2, 'x') + BV.uatom(2, 'y')).with_output('z')\n\n# Create distribution over bits.\npcirc = C.pcirc(adder)                                         \\\n         .randomize({'x': {0: 1/6, 1: 2/6, 2: 3/6}})\npcirc >>= incr\n```\n\nor\n\n```python\ninc_x = C.pcirc(BV.uatom(2, 'x') + 1)          \\\n         .randomize({'x': {0: 1/3, 2: 2/3}})                      # Pr(x=2) = 2/3\n\ninc_y = C.pcirc(BV.uatom(3, 'y') + 1)          \\\n         .randomize({'y': {0: 1/3, 5: 1/3, 3: 1/3}})\n\ninc_xy = inc_x | inc_y  #\n```\n\nSimilarly, `unroll`, `loopback` are also implemented.\n\n**note** `unroll` combines all coin flips into a *single* input in\ntemporal order.\n\n\n## Finite Functions\n\n`py-aiger-coins` also works well with the `py-aiger-discrete` API for\nworking with arbitrary functions over finite sets. For example:\n\n```python\nfrom bidict import bidict  # `pip install bidict`\n\n# Create encoder/decoder for dice.\nlookup = bidict({0: '⚀', 1: '⚁', 2: '⚂', 3: '⚃'})     # invertable dictionary.\nencoder = aiger_discrete.Encoding(decode=lookup.get, encode=lookup.inv.get)\n\n# Represent dice with a 2 bit vector.\nexpr1 = BV.uatom(2, '🎲')\n\n# Add encoded dice to x. Because why not.\nexpr2 = BV.uatom(2, 'x') + expr1\nfunc = aiger_discrete.from_aigbv(\n    expr2.aigbv, input_encodings={'🎲': encoder}\n)\n\n# Create distribution over bits.\ncirc = C.pcirc(func) \\\n        .randomize({'🎲': {'⚀': 1/6, '⚁': 2/6, '⚂': 3/6}})\n\nassert circ.inputs == {'x'}\nassert circ.outputs == {expr2.output}\nassert circ.coin_biases == (Fraction(1, 3), Fraction(1, 2))\n```\n",
    'author': 'Marcell Vazquez-Chanlatte',
    'author_email': 'mvc@linux.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mvcisback/py-aiger-coins',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
